import sympy as sp
from mpi4py import MPI
import pytest
from shenfun import *

comm = MPI.COMM_WORLD

def test_lagrangian_particles():
    N = (20, 20)
    F0 = FunctionSpace(N[0], 'F', dtype='D', domain=(0., 1.))
    F1 = FunctionSpace(N[1], 'F', dtype='d', domain=(0., 1.))
    T = TensorProductSpace(comm, (F0, F1))
    TV = VectorSpace(T)

    x, y = sp.symbols("x,y")
    psi = 1./np.pi*sp.sin(np.pi*x)**2*sp.sin(np.pi*y)**2 # Streamfunction
    ux = -psi.diff(y, 1)
    uy = psi.diff(x, 1)

    uxl = sp.lambdify((x, y), ux, 'numpy')
    uyl = sp.lambdify((x, y), uy, 'numpy')
    X = T.local_mesh(True)
    u = Array(T, buffer=uxl(X[0], X[1]))
    v = Array(T, buffer=uyl(X[0], X[1]))
    uv = Function(TV)
    uv[0] = T.forward(u, uv[0])
    uv[1] = T.forward(v, uv[1])

    points = np.array([[0.5], [0.75]])

    # Create LagrangianParticles instance with given points
    dt = 0.01
    lp = LagrangianParticles(points, dt, uv)
    for i in range(4):
        lp.step()

    assert np.allclose(lp.x, np.array([[0.53986228], [0.74811753]]), 1e-6)
    assert np.allclose(lp.up, np.array([[0.99115526], [-0.09409196]]), 1e-6)
    T.destroy()

if __name__ == '__main__':
    test_lagrangian_particles()



#cython: boundscheck=False
#cython: wraparound=False
#cython: language_level=3

from libc.math cimport sin, cos, exp, sqrt, lgamma, M_PI, M_2_PI
from libcpp.vector cimport vector
from libc.stdlib cimport malloc, free
import array
import numpy as np
cimport cython
cimport numpy as np
import cython
from scipy.special import gammaln
np.import_array()

ctypedef fused T:
    double
    complex

ctypedef void (*funv)(T* const, T*, int, int , void* const)

ctypedef bint bool

cpdef double normf(double[::1] u):
    cdef:
        int i
        int N = u.shape[0]
        double s = 0
    for i in range(N):
        s += u[i]*u[i]
    s = sqrt(s)
    return s

cdef double norm(double* u, int N):
    cdef:
        int i
        double s = 0
    for i in range(N):
        s += u[i]*u[i]
    s = sqrt(s)
    return s

cdef double sum(double* u, int N):
    cdef:
        int i
        double s = 0
    for i in range(N):
        s += u[i]
    return s

cdef void IterAllButAxis(funv f, np.ndarray[T, ndim=1] input_array, np.ndarray[T, ndim=1] output_array, int st, int N, int axis, tuple shapein, tuple shapeout, void* const data):
    cdef:
        np.flatiter ita = np.PyArray_IterAllButAxis(np.PyArray_Reshape(input_array, shapein), &axis)
        np.flatiter ito = np.PyArray_IterAllButAxis(np.PyArray_Reshape(output_array, shapeout), &axis)

    while np.PyArray_ITER_NOTDONE(ita):
        f(<T* const>np.PyArray_ITER_DATA(ita), <T*>np.PyArray_ITER_DATA(ito), st, N, data)
        np.PyArray_ITER_NEXT(ita)
        np.PyArray_ITER_NEXT(ito)

cdef void _matvec(double* A, T* x, T* b, int m, int n, int transpose):
    cdef:
        int i, j, k
        T s
        double* a
    if transpose == 0:
        for i in range(m):
            s = 0.0
            a = &A[n*i]
            for j in range(n):
                s += a[j]*x[j]
            b[i] = s
    else:
        for j in range(n):
            b[j] = 0
        for i in range(m):
            s = x[i]
            a = &A[n*i]
            for j in range(n):
                b[j] += s*a[j]

cdef void _matvecadd(double* A, T* x, T* b, int m, int n, int transpose):
    cdef:
        int i, j, k
        T s
        double* a
    if transpose == 0:
        for i in range(m):
            s = 0.0
            a = &A[n*i]
            for j in range(n):
                s += a[j]*x[j]
            b[i] += s
    else:
        for i in range(m):
            s = x[i]
            a = &A[n*i]
            for j in range(n):
                b[j] += s*a[j]


cdef void _matvectri(double* A, T* x, T* b, int m, int n, int lda, int lower):
    cdef:
        T* xp
        double* ap
        T s
        int i, j

    if lower == 0:
        for i in range(m):
            s = 0.0
            xp = &x[0]
            ap = &A[i*lda]
            for j in range(i+1):
                s += ap[j]*xp[j]
            b[i] += s
    else:
        for i in range(m):
            s = 0.0
            xp = &x[i]
            ap = &A[i*lda+i]
            for j in range(n-i):
                s += ap[j]*xp[j]
            b[i] += s

ctypedef struct R0:
    double* a
    double* x
    double* Lnm
    double* Ln
    double* Lnp
    int No
    int M
    int Mx

cpdef np.ndarray restricted_product(L, np.ndarray input_array, np.ndarray output_array, double[::1] x, int i0, int i1, int a0, int axis, np.ndarray[double, ndim=2] a):
    cdef:
        int n = input_array.ndim
        int st, N
        int M = output_array.shape[axis]
        str dtype = input_array.dtype.char
        tuple shapein
        tuple shapeout = np.shape(output_array)
        np.ndarray[double, ndim=1] Lnm = L.evaluate_basis(x[i0:i1], i=a0)
        np.ndarray[double, ndim=1] Ln = L.evaluate_basis(x[i0:i1], i=a0+1)
        np.ndarray[double, ndim=1] Lnp = L.evaluate_basis(x[i0:i1], i=a0+2)
        np.ndarray[double, ndim=2] ax = a[:, slice(a0, None)].copy()
        R0 r0 = R0(&ax[0, 0],
                   &x[i0], &Lnm[0], &Ln[0], &Lnp[0],
                   M, ax.shape[0], ax.shape[1])
    sl = [slice(None)]*n
    sl[axis] = slice(i0, i1)
    input = input_array[tuple(sl)]
    if input.flags['C_CONTIGUOUS'] is False:
        input = input.copy()
    shapein = np.shape(input)
    N = input.shape[axis]
    st = input.strides[axis]//input.itemsize
    if dtype == 'd':
        IterAllButAxis[double](_restricted_product_ptr, np.PyArray_Ravel(input, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shapein, shapeout, &r0)
    elif dtype == 'D':
        IterAllButAxis[complex](_restricted_product_ptr, np.PyArray_Ravel(input, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shapein, shapeout, &r0)
    else:
        raise NotImplementedError
    return output_array

ctypedef struct S0:
    double* xj
    double* wj
    double* a
    int Nx
    int M
    int Mx

cpdef scalar_product(np.ndarray input_array, np.ndarray output_array, double[::1] x, double[::1] w, int axis, double[:, ::1] a):
    cdef:
        int n = input_array.ndim
        int st = input_array.strides[axis]//input_array.itemsize
        int N = input_array.shape[axis]
        str dtype = input_array.dtype.char
        tuple shapein = np.shape(input_array)
        tuple shapeout = np.shape(output_array)
        S0 s0 = S0(&x[0], &w[0], &a[0, 0], output_array.shape[axis], a.shape[0], a.shape[1])

    if dtype == 'd':
        IterAllButAxis[double](_scalar_product_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shapein, shapeout, &s0)
    elif dtype == 'D':
        IterAllButAxis[complex](_scalar_product_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shapein, shapeout, &s0)
    else:
        raise NotImplementedError

ctypedef struct E0:
    double* x
    double* a
    int Nx
    int M
    int Mx

cpdef evaluate_expansion_all(np.ndarray input_array, np.ndarray output_array, double[::1] x, int axis, double[:, ::1] a):
    cdef:
        int n = input_array.ndim
        int st = input_array.strides[axis]//input_array.itemsize
        int N = input_array.shape[axis]
        str dtype = input_array.dtype.char
        tuple shapein = np.shape(input_array)
        tuple shapeout = np.shape(output_array)
        E0 e0 = E0(&x[0], &a[0, 0], output_array.shape[axis], a.shape[0], a.shape[1])

    if not input_array.flags['C_CONTIGUOUS']:
        input_array = input_array.copy()
    if dtype == 'd':
        IterAllButAxis[double](_evaluate_expansion_all_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shapein, shapeout, &e0)
    elif dtype == 'D':
        IterAllButAxis[complex](_evaluate_expansion_all_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shapein, shapeout, &e0)
    else:
        raise NotImplementedError

@cython.cdivision(True)
cdef void _evaluate_expansion_all_ptr(T* const ui,
                                      T* uo,
                                      int st,
                                      int N,
                                      void* const data):
    cdef:
        int i, j
        T s
        E0* e0 = <E0*> data
        int Nx = e0.Nx
        int Mx = e0.Mx
        int M = e0.M
        double s1, s2, a00
        double* xj = e0.x
        double* a = e0.a
        double* anm = &a[0]
        double* anp
        double* ann
        double *an = <double*>malloc(Mx*sizeof(double))
        double *Lnm = <double*>malloc(Nx*sizeof(double))
        double *Ln = <double*>malloc(Nx*sizeof(double))
        double *Lnp = <double*>malloc(Nx*sizeof(double))

    if M == 2:
        for i in range(Mx):
            an[i] = 0
        ann = &an[0]
        anp = &a[Mx]
    else:
        ann = &a[Mx]
        anp = &a[2*Mx]
    for i in range(Nx):
        Lnm[i] = 1
        Ln[i] = (xj[i]-ann[0])/anm[0]
        uo[i*st] = 0
        Lnp[i] = (xj[i]-ann[1])/anm[1]*Ln[i] - anp[1]/anm[1]*Lnm[i]
    for i in range(N):
        s1 = 1/anm[i+2]
        s2 = anp[i+2]/anm[i+2]
        a00 = ann[i+2]
        for j in range(Nx):
            uo[j*st] += Lnm[j]*ui[i*st]
            Lnm[j] = Ln[j]
            Ln[j] = Lnp[j]
            Lnp[j] = s1*(xj[j]-a00)*Ln[j] - s2*Lnm[j]
    free(an)
    free(Lnm)
    free(Ln)
    free(Lnp)

@cython.cdivision(True)
cdef void _scalar_product_ptr(T* const ui,
                              T* uo,
                              int st,
                              int N,
                              void* const data):
    cdef:
        int i, j
        T s
        double s1, s2, a00
        S0* s0 = <S0*> data
        int M = s0.M
        int Mx = s0.Mx
        int Nx = s0.Nx
        double* xj = s0.xj
        double* wj = s0.wj
        double* a = s0.a
        double* anm = &a[0]
        double* ann
        double* anp
        double *an = <double*>malloc(Mx*sizeof(double))
        double *Lnm = <double*>malloc(Nx*sizeof(double))
        double *Ln = <double*>malloc(Nx*sizeof(double))
        double *Lnp = <double*>malloc(Nx*sizeof(double))

    if M == 2:
        for i in range(Mx):
            an[i] = 0
        ann = &an[0]
        anp = &a[Mx]
    else:
        ann = &a[Mx]
        anp = &a[2*Mx]

    for i in range(Nx):
        Lnm[i] = 1
        Ln[i] = (xj[i]-ann[0])/anm[0]
        uo[i*st] = 0
        Lnp[i] = (xj[i]-ann[1])/anm[1]*Ln[i] - anp[1]/anm[1]*Lnm[i]

    for i in range(N):
        s1 = 1/anm[i+2]
        s2 = anp[i+2]/anm[i+2]
        a00 = ann[i+2]
        s = 0.0
        for j in range(Nx):
            s += Lnm[j]*wj[j]*ui[j*st]
            Lnm[j] = Ln[j]
            Ln[j] = Lnp[j]
            Lnp[j] = s1*(xj[j]-a00)*Ln[j] - s2*Lnm[j]
        uo[i*st] = s
    free(an)
    free(Lnm)
    free(Ln)
    free(Lnp)


@cython.cdivision(True)
cdef void _restricted_product_ptr(T* const input_array,
                                  T* output_array,
                                  int st,
                                  int N,
                                  void* const data):

    cdef:
        int k, kp, i
        double a00
        T s
        R0* r0 = <R0*>data
        double* xi = r0.x
        double* a = r0.a
        int No = r0.No
        int M = r0.M
        int Mx = r0.Mx
        double* anm = &a[0]
        double* ann
        double* anp
        double *an = <double*>malloc(Mx*sizeof(double))
        double *Lnm = <double*>malloc(N*sizeof(double))
        double *Ln = <double*>malloc(N*sizeof(double))
        double *Lnp = <double*>malloc(N*sizeof(double))

    if M == 2:
        for i in range(Mx):
            an[i] = 0
        ann = &an[0]
        anp = &a[Mx]
    else:
        ann = &a[Mx]
        anp = &a[2*Mx]
    for i in range(N):
        Lnm[i] = r0.Lnm[i]
        Ln[i] = r0.Ln[i]
        Lnp[i] = r0.Lnp[i]
    for k in range(No):
        s1 = 1/anm[k+2]
        s2 = anp[k+2]/anm[k+2]
        a00 = ann[k+2]
        s = 0.0
        for i in range(N):
            s += Lnm[i]*input_array[i*st]
            Lnm[i] = Ln[i]
            Ln[i] = Lnp[i]
            Lnp[i] = s1*(xi[i]-a00)*Ln[i] - s2*Lnm[i]
        output_array[k*st] = s
    free(an)
    free(Lnm)
    free(Ln)
    free(Lnp)

cdef int get_h(int level, long* D, int N):
    cdef:
        int i
        int h = 1
    if level < N-1:
        for i in range(level+1, N):
            h *= D[i]
    return h

cdef int get_number_of_blocks(int level, long* D, int N):
    cdef:
        int s = 1
        int nd = 1
        int i, j
    if level > 0:
        for i in range(level):
            nd = 1
            for j in range(i+1):
                nd *= D[level-j-1]
            s += nd
    return s

cdef int get_number_of_submatrices(long* D, int L):
    cdef:
        int Ns, level
    Ns = 0
    for level in range(L):
        Ns += get_number_of_blocks(level, D, L)*D[level]*(D[level]+1)//2
    return Ns

cdef void get_ij(int* ij, int level, int block, int diags, int h, int s, long* D, int L):
    cdef:
        int i, i0, j0
    ij[0] = block*D[level]*s*get_h(level, D, L)
    ij[1] = diags+ij[0]
    for i in range(level+1, L):
        ij[1] += s*get_h(i, D, L)

cdef double chebvalS(double *x, double* c, int M):
    cdef:
        double x2, c0, c1, tmp
        int i
    x2 = 2*x[0]
    c0 = c[M-2]
    c1 = c[M-1]
    for i in range(3, M + 1):
        tmp = c0
        c0 = c[M-i] - c1
        c1 = tmp + c1*x2
    return c0+c1*x[0]

cdef void chebvalC(double* x, int N, double* c, int M, double* c0):
    cdef:
        double* x2 = <double*>malloc(N*sizeof(double))
        double* c1 = <double*>malloc(N*sizeof(double))
        int i, j
        double tmp

    for j in range(N):
        x2[j] = 2*x[j]
        c0[j] = c[M-2]
        c1[j] = c[M-1]
    for i in range(3, M + 1):
        for j in range(N):
            tmp = c0[j]
            c0[j] = c[M-i] - c1[j]
            c1[j] = tmp + c1[j]*x2[j]
    for j in range(N):
        c0[j] += c1[j]*x[j]
    free(x2)
    free(c1)

cpdef np.ndarray[double, ndim=1] chebval(np.ndarray[double, ndim=1] x, np.ndarray[double, ndim=1] u):
    cdef:
        int N = x.shape[0]
        int M = u.shape[0]
        np.ndarray[double, ndim=1] c = u.copy()
        np.ndarray[double, ndim=1] x2 = x.copy()
        np.ndarray[double, ndim=1] c0 = np.zeros_like(x)
        np.ndarray[double, ndim=1] c1 = np.zeros_like(x)
        int i, j
        double tmp

    if M == 1:
        c0[:] = c[0]
        c1[:] = 0
    elif M == 2:
        c0[:] = c[0]
        c1[:] = c[1]
    else:
        for i in range(N):
            x2[i] = 2*x[i]
        c0[:] = c[M-2]
        c1[:] = c[M-1]
        for i in range(3, M + 1):
            for j in range(N):
                tmp = c0[j]
                c0[j] = c[M-i] - c1[j]
                c1[j] = tmp + c1[j]*x2[j]
    for j in range(N):
        c0[j] += c1[j]*x[j]
    return c0

def Lambda(np.ndarray x):
    """Return

    .. math::

            \Lambda(x) = \frac{\Gamma(x+\frac{1}{2})}{\Gamma(x+1)}

    Parameters
    ----------
    x : array of floats
    """
    cdef:
        int ndim = np.PyArray_NDIM(x)
        np.ndarray[double, ndim=1] a = _Lambda(np.PyArray_Ravel(x, np.NPY_CORDER))
    return np.PyArray_Reshape(a, np.shape(x))

@cython.cdivision(True)
cpdef np.ndarray[double, ndim=1] _Lambda(double[::1] z):
    cdef:
        int N = z.shape[0]
        int i = 0
        double x
        np.ndarray[double, ndim=1] a = np.empty(N)
        double[8] a0 = [9.9688251374224490e-01,
                        -3.1149502185860763e-03,
                        2.5548605043159494e-06,
                        1.8781800445057731e-08,
                        -4.0437919461099256e-11,
                        -9.0003384202201278e-13,
                        3.1032782098712292e-15,
                        1.0511830721865363e-16]

    for i in range(N):
        if z[i] > 20:
            x = -1+40/z[i]
            a[i] = chebvalS(&x, a0, 8)
            a[i] /= sqrt(z[i])
        else:
            a[i] = exp(lgamma(z[i]+0.5) - lgamma(z[i]+1))
    return a

ctypedef struct DL:
    double* a
    int transpose

@cython.cdivision(True)
cpdef _leg2cheb(np.ndarray[T, ndim=1] ui, np.ndarray[T, ndim=1] uo, double[::1] a, int trans):
    cdef:
        DL d = DL(&a[0], trans)
        int st = ui.strides[0]
        int sz = <int>ui.itemsize
        int N = ui.shape[0]
    _leg2cheb_ptr[T](&ui[0], &uo[0], st//sz, N, &d)

ctypedef struct DI:
    double* a
    double* d

cpdef _cheb2leg(T[::1] ui, T[::1] uo, double[::1] dn, double[::1] a):
    cdef DI di = DI(&a[0], &dn[0])
    _cheb2leg_ptr[T](&ui[0], &uo[0], ui.strides[0]/ui.itemsize, ui.shape[0], &di)

cpdef np.ndarray cheb2leg(np.ndarray input_array, np.ndarray output_array, int axis=0):
    cdef:
        int N = input_array.shape[axis]
        int st = input_array.strides[axis]//input_array.itemsize
        int Ne = N//2+N%2
        str dtype = input_array.dtype.char
        np.ndarray[double, ndim=1] k = np.arange(N, dtype='d')
        np.ndarray[double, ndim=1] a = np.zeros(N)
        np.ndarray[double, ndim=1] dn = np.zeros(Ne)
        DI d = DI(&a[0], &dn[0])
        tuple shape = np.shape(input_array)

    k[0] = 1
    dn[:] = _Lambda((k[::2]-2)/2)/k[::2]
    a[:] = 1/(2*_Lambda(k)*k*(k+0.5))
    a[0] = 2/np.sqrt(np.pi)
    output_array.fill(0)
    if dtype == 'd':
        IterAllButAxis[double](_cheb2leg_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d)
    elif dtype == 'D':
        IterAllButAxis[complex](_cheb2leg_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d)
    else:
        raise NotImplementedError
    return output_array

cpdef np.ndarray leg2cheb(np.ndarray input_array, np.ndarray output_array, int axis=0, int transpose=0):
    cdef:
        int N = input_array.shape[axis]
        int st = input_array.strides[axis]//input_array.itemsize
        str dtype = input_array.dtype.char
        np.ndarray[double, ndim=1] a = Lambda(np.arange(N, dtype=np.float64))
        DL d = DL(&a[0], transpose)
        tuple shape = np.shape(input_array)

    output_array.fill(0)
    if dtype == 'd':
        IterAllButAxis[double](_leg2cheb_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d)
    elif dtype == 'D':
        IterAllButAxis[complex](_leg2cheb_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d)
    else:
        raise NotImplementedError
    return output_array

@cython.cdivision(True)
cdef void _leg2cheb_ptr(T* const c,
                        T* v,
                        int st,
                        int N,
                        void* const data):
    cdef:
        int n, i, ii
        DL* d = <DL*>data
        double* a = d.a
        double* ap
        T* cp
        int transpose = d.transpose
        T *cv = <T *> malloc(N*transpose*sizeof(T))

    if transpose == 0:
        for n in range(0, N, 2):
            ap = &a[n//2]
            cp = &c[n*st]
            ii = 0
            for i in range(0, N-n):
                v[ii] += ap[0]*ap[i]*cp[ii]
                ii += st
        v[0] *= 0.5
        ii = 0
        for n in range(0, N):
            v[ii] *= M_2_PI
            ii += st
    else:
        ii = 0
        for i in range(0, N):
            cv[i] = M_2_PI*c[ii]
            ii += st
        cv[0] *= 0.5
        for n in range(0, N, 2):
            ap = &a[n//2]
            ii = n*st
            for i in range(0, N-n):
                v[ii] += ap[0]*ap[i]*cv[i]
                ii += st
    free(cv)

@cython.cdivision(True)
cdef void _cheb2leg_ptr(T* const v,
                        T* c,
                        int st,
                        int N,
                        void* const data):
    cdef:
        int n, i, ii
        double SPI = sqrt(M_PI)
        DI* di = <DI*>data
        double* a = di.a
        double* d = di.d
        T* vn = <T*>malloc(N*sizeof(T))
    vn[0] = v[0]
    ii = st
    for i in range(1, N):
        vn[i] = v[ii]*i
        ii += st
    ii = 0
    for n in range(N):
        c[ii] = SPI*a[n]*vn[n]
        ii += st
    for n in range(2, N, 2):
        ii = 0
        for i in range(0, N-n):
            c[ii] -= d[n//2]*a[n//2+i]*vn[n+i]
            ii += st
    ii = 0
    for n in range(N):
        c[ii] *= (n+0.5)
        ii += st
    free(vn)

ctypedef struct DCN:
    int Nn
    double* A
    double* T
    double* Th
    double* ThT
    long* Nk
    long* D
    long* Mmin
    long* uD
    long* cuD
    size_t Lh
    size_t L
    size_t s
    size_t diags
    bool trans

cdef size_t find_index(long* v, size_t u, size_t N):
    cdef size_t i = 1
    if v[0] == u:
        return 0
    while v[i] != u:
        i += 1
    return i

@cython.cdivision(True)
cpdef void FMMcheb(np.ndarray input_array, np.ndarray output_array, int axis, int Nn, double[::1] A, np.ndarray[long, ndim=1] Nk, double[:, :, ::1] T, double[:, :, ::1] Th, double[:, :, ::1] ThT, long[::1] D, long[::1] Mmin, int s, int diags, int trans):
    cdef:
        np.ndarray[long, ndim=1] uD = np.hstack((0, np.unique(D)))
        np.ndarray[long, ndim=1] cuD = np.cumsum(uD)
        DCN dc = DCN(Nn, &A[0], &T[0, 0, 0], &Th[0, 0, 0], &ThT[0, 0, 0], &Nk[0], &D[0], &Mmin[0], &uD[0], &cuD[0], len(uD), len(D), s, diags, trans)
        int st = input_array.strides[axis]//input_array.itemsize
        int N = input_array.shape[axis]
        dtype = input_array.dtype.char
        tuple shape = np.shape(input_array)
    if dtype == 'd':
        IterAllButAxis[double](_FMMcheb_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &dc)
    elif dtype == 'D':
        IterAllButAxis[complex](_FMMcheb_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &dc)
    else:
        raise NotImplementedError

@cython.cdivision(True)
cdef void _divmod(size_t a, size_t b, size_t* i0, size_t* j0):
    i0[0] = a // b;
    j0[0] = a % b;

@cython.cdivision(True)
cdef void _FMMcheb_ptr(T* u,
                       T* v,
                       int st,
                       int N,
                       void* data):
    cdef:
        size_t ik, i0, j0, j1, i, M, odd, Nc
        size_t s0, s1, M0, Mmax, block, level, p, q, h, b0, p0, q0
        int ij[2]
        DCN* dc = <DCN*>data
        double* A = dc.A
        double* TT = dc.T
        long* Nk = dc.Nk
        long* D = dc.D
        long* Mmin = dc.Mmin
        double* Th = dc.Th
        double* ThT = dc.ThT
        long* uD = dc.uD
        long* cuD = dc.cuD
        size_t Lh = dc.Lh
        size_t Nn = dc.Nn
        size_t L = dc.L
        size_t s = dc.s
        size_t diags = dc.diags
        bool trans = dc.trans
        T* cia = <T*>malloc(Nn//2*sizeof(T))
        T* coa = <T*>malloc(Nn//2*sizeof(T))
        T** wk = <T**>malloc(L*sizeof(T*))
        T** ck = <T**>malloc(L*sizeof(T*))
        T* c0
        T* c1
        T* w0
        T* w1

    Mmax = 0
    for i in range(L):
        Mmax = max(Mmin[i], Mmax)

    for level in range(L):
        ck[level] = <T*>malloc(get_number_of_blocks(level, D, L)*D[level]*Mmax*sizeof(T))
        wk[level] = <T*>malloc(get_number_of_blocks(level, D, L)*D[level]*Mmax*sizeof(T))
    jj = 0 if level == L-1 else D[level+1]
    if trans == 0:
        for odd in range(2):
            for i in range(Nn//2):
                cia[i] = 0
                coa[i] = 0
            for i in range(N//2+(N%2)*(1-odd)):
                cia[i] = u[(2*i+odd)*st]
            for level in range(L):
                for i in range(get_number_of_blocks(level, D, L)*D[level]*Mmax):
                    wk[level][i] = 0.0
                    ck[level][i] = 0.0
            ik = 0
            Nc = 0
            for level in range(L-1, -1, -1):
                M0 = Mmin[level]
                h = s*get_h(level, D, L)
                s0 = find_index(uD, D[level], Lh)-1
                s1 = cuD[s0]

                for block in range(get_number_of_blocks(level, D, L)):
                    get_ij(&ij[0], level, block, diags, h, s, D, L)
                    c0 = &ck[level][block*D[level]*Mmax]
                    w0 = &wk[level][block*D[level]*Mmax]

                    for q in range(D[level]):
                        if level == L-1:
                            _matvec(&TT[odd*s*M0], &cia[ij[1]+(q+1)*s], &w0[q*Mmax], s, M0, 1)
                        if level > 0 and block > 0:
                            _divmod(block-1, D[level-1], &b0, &q0)
                            _matvectri(&Th[(s1+q)*Mmax*Mmax], &w0[q*Mmax], &wk[level-1][(b0*D[level-1]+q0)*Mmax], M0, M0, Mmax, 0)

                        for p in range(q+1):
                            M = Nk[ik]
                            _matvecadd(&A[Nc], &w0[q*Mmax], &c0[p*Mmax], M, M, 0)
                            Nc += M*M
                            ik += 1

            for level in range(L-1):
                c0 = ck[level]
                c1 = ck[level+1]
                M0 = Mmin[level]
                j1 = 0
                s0 = find_index(uD, D[level+1], Lh)-1
                s1 = cuD[s0]
                for block in range(get_number_of_blocks(level+1, D, L)-1):
                    for p in range(D[level+1]):
                        _matvectri(&ThT[(s1+p)*Mmax*Mmax], &c0[block*Mmax], &c1[j1*Mmax], M0, M0, Mmax, 1)
                        j1 += 1

            M0 = Mmin[L-1]
            level = L-1
            for block in range(get_number_of_blocks(level, D, L)):
                get_ij(&ij[0], level, block, diags, s, s, D, L)
                c0 = &ck[level][block*D[level]*Mmax]
                for p in range(D[level]):
                    _matvecadd(&TT[odd*s*M0], &c0[p*Mmax], &coa[ij[0]+p*s], s, M0, 0)

            for i in range(N//2+(N%2)*(1-odd)):
                v[(2*i+odd)*st] = coa[i]

    else:
        for odd in range(2):
            for i in range(Nn//2):
                cia[i] = 0
                coa[i] = 0
            for i in range(N//2+(N%2)*(1-odd)):
                cia[i] = u[(2*i+odd)*st]
            for level in range(L):
                for i in range(get_number_of_blocks(level, D, L)*D[level]*Mmax):
                    wk[level][i] = 0.0
                    ck[level][i] = 0.0
            ik = 0
            Nc = 0
            for level in range(L-1, -1, -1):
                M0 = Mmin[level]
                h = s*get_h(level, D, L)
                w0 = wk[level]
                for block in range(get_number_of_blocks(level, D, L)):
                    get_ij(&ij[0], level, block, diags, h, s, D, L)
                    i0 = ij[1]
                    j0 = ij[0]
                    c0 = &ck[level][block*D[level]*Mmax]
                    w0 = &wk[level][block*D[level]*Mmax]
                    s0 = find_index(uD, D[level], Lh)-1
                    s1 = cuD[s0]
                    _divmod(block, D[level], &b0, &q0)
                    for p in range(D[level]):
                        for q in range(p+1):
                            if q == p:
                                if level == L-1:
                                    _matvecadd(&TT[odd*s*M0], &cia[j0+q*s], &w0[q*Mmax], s, M0, 1)
                                if level > 0 and block < get_number_of_blocks(level, D, L)-1:
                                    _matvectri(&Th[(s1+q)*Mmax*Mmax], &w0[q*Mmax], &wk[level-1][(b0*D[level]+q0)*Mmax], M0, M0, Mmax, 0)
                            M = Nk[ik]
                            _matvecadd(&A[Nc], &w0[q*Mmax], &c0[p*Mmax], M, M, 1)
                            Nc += M*M
                            ik += 1

            for level in range(L-1):
                c0 = ck[level]
                c1 = ck[level+1]
                M0 = Mmin[level]
                s0 = find_index(uD, D[level+1], Lh)-1
                s1 = cuD[s0]
                for block in range(1, get_number_of_blocks(level+1, D, L)):
                    for p in range(D[level+1]):
                        _divmod(block-1, D[level], &b0, &p0)
                        _matvectri(&ThT[(s1+p)*Mmax*Mmax], &c0[(b0*D[level]+p0)*Mmax], &c1[(block*D[level+1]+p)*Mmax], M0, M0, Mmax, 1)

            M0 = Mmin[L-1]
            level = L-1
            for block in range(get_number_of_blocks(level, D, L)):
                get_ij(&ij[0], level, block, diags, s, s, D, L)
                i0 = ij[1]
                j0 = ij[0]
                c0 = &ck[level][block*D[level]*Mmax]
                for p in range(D[level]):
                    _matvecadd(&TT[odd*s*M0], &c0[p*Mmax], &coa[i0+(p+1)*s], s, M0, 0)

            for i in range(N//2+(N%2)*(1-odd)):
                v[(2*i+odd)*st] = coa[i]

    free(cia)
    free(coa)
    for level in range(L):
        free(wk[level])
        free(ck[level])
    free(wk)
    free(ck)


ctypedef struct D1:
    double* a
    int n0
    int trans


cpdef void FMMdirect1(np.ndarray input_array, np.ndarray output_array, int axis, double[::1] a, int n0, int trans):
    cdef:
        D1 d1 = D1(&a[0], n0, trans)
        int st = input_array.strides[axis]//input_array.itemsize
        int N = input_array.shape[axis]
        dtype = input_array.dtype.char
        tuple shape = np.shape(input_array)

    if dtype == 'd':
        IterAllButAxis[double](_FMMdirect1_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d1)
    elif dtype == 'D':
        IterAllButAxis[complex](_FMMdirect1_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d1)
    else:
        raise NotImplementedError

@cython.cdivision(True)
cdef void _FMMdirect1_ptr(T* const u,
                          T* v,
                          int st,
                          int N,
                          void* const data):
    cdef:
        D1* d1 = <D1*>data
        int n0 = d1.n0
        int trans = d1.trans
        double* a = d1.a
        double* ap
        T* cp
        int n, i, ii

    if trans == 0:
        for n in range(0, 2*n0, 2):
            ap = &a[n//2]
            cp = &u[n*st]
            ii = 0
            for i in range(0, N-n):
                v[ii] += ap[0]*ap[i]*cp[ii]
                ii += st
    else:
        for n in range(0, 2*n0, 2):
            ap = &a[n//2]
            cp = &v[n*st]
            ii = 0
            for i in range(0, N-n):
                cp[ii] += ap[0]*ap[i]*u[ii]
                ii += st

ctypedef struct D2:
    double* a
    int h
    int Nd
    int n0
    int trans

cpdef void FMMdirect2(np.ndarray input_array, np.ndarray output_array, int axis, double[::1] a, int h, int Nd, int n0, int trans):
    cdef:
        D2 d2 = D2(&a[0], h, Nd, n0, trans)
        int st = input_array.strides[axis]//input_array.itemsize
        int N = input_array.shape[axis]
        dtype = input_array.dtype.char
        tuple shape= np.shape(input_array)

    if dtype == 'd':
        IterAllButAxis[double](_FMMdirect2_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d2)
    elif dtype == 'D':
        IterAllButAxis[complex](_FMMdirect2_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d2)
    else:
        raise NotImplementedError

@cython.cdivision(True)
cdef void _FMMdirect2_ptr(T* const u,
                          T* v,
                          int st,
                          int N,
                          void* const data):
    cdef:
        int n, i, j, k, d, i0, j0, n1, ii, Nm
        D2* d2 = <D2*>data
        double* a = d2.a
        int h = d2.h
        int Nd = d2.Nd
        int n0 = d2.n0
        int trans = d2.trans
        T* vp
        T* up
        double* ap

    for k in range(Nd):
        i0 = k*h
        j0 = n0+i0
        for n in range(0, h, 2):
            ap = &a[i0+(n0+n)//2]
            Nm = min(N-(j0+n), h-n)
            n1 = (n+n0)//2
            if trans == 0:
                vp = &v[i0*st]
                up = &u[(j0+n)*st]
                ii = 0
                for j in range(Nm):
                    vp[ii] += a[n1]*ap[j]*up[ii]
                    ii += st
            else:
                vp = &v[(j0+n)*st]
                up = &u[i0*st]
                ii = 0
                for j in range(Nm):
                    vp[ii] += a[n1]*ap[j]*up[ii]
                    ii += st


ctypedef struct D3:
    double* dn
    double* a
    int h
    int Nd
    int n0

cpdef void FMMdirect3(np.ndarray input_array, np.ndarray output_array, int axis, double[::1] dn, double[::1] a, int h, int Nd, int n0):
    cdef:
        int st = input_array.strides[axis]//input_array.itemsize
        int N = input_array.shape[axis]
        dtype = input_array.dtype.char
        D3 d3 = D3(&dn[0], &a[0], h, Nd, n0)
        tuple shape = np.shape(input_array)

    if dtype == 'd':
        IterAllButAxis[double](_FMMdirect3_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d3)
    elif dtype == 'D':
        IterAllButAxis[complex](_FMMdirect3_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d3)
    else:
        raise NotImplementedError


cdef void _FMMdirect3_ptr(T* const u,
                          T* v,
                          int st,
                          int N,
                          void* const data):
    cdef:
        int n, i, j, k, d, n1, i0, j0, ii, Nm
        D3* d3 = <D3*>data
        double* a = d3.a
        double* dn = d3.dn
        int h = d3.h
        int Nd = d3.Nd
        int n0 = d3.n0
        T* up
        T* vp
        double* dp
        double* ap

    for k in range(Nd):
        i0 = k*h
        j0 = n0+i0
        vp = &v[i0*st]
        for n in range(0, h, 2):
            n1 = (n+n0)//2
            Nm = min(N-(j0+n), h-n)
            ap = &a[i0+n1]
            up = &u[(j0+n)*st]
            ii = 0
            for j in range(Nm):
                vp[ii] -= dn[n1]*ap[j]*up[ii]
                ii += st

ctypedef struct D4:
    double* dn
    double* a
    int n0

cdef void _FMMdirect4_ptr(T* const u,
                          T* v,
                          int st,
                          int N,
                          void* const data):
    cdef:
        int n, i, ii
        D4* d4 = <D4*>data
        double* a = d4.a
        double* dn = d4.dn
        int n0 = d4.n0
        double* dp
        double* ap
        T* up
        double SPI = sqrt(M_PI)

    for i in range(N):
        v[i*st] += SPI*a[i]*u[i*st]
    for n in range(2, 2*n0, 2):
        ap = &a[n//2]
        up = &u[n*st]
        ii = 0
        for i in range(N-n):
            v[ii] -= dn[n//2]*ap[i]*up[ii]
            ii += st

cpdef void FMMdirect4(np.ndarray input_array, np.ndarray output_array, int axis, double[::1] dn, double[::1] a, int n0):
    cdef:
        int st = input_array.strides[axis]//input_array.itemsize
        int N = input_array.shape[axis]
        dtype = input_array.dtype.char
        D4 d4 = D4(&dn[0], &a[0], n0)
        tuple shape = np.shape(input_array)

    if dtype == 'd':
        IterAllButAxis[double](_FMMdirect4_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d4)
    elif dtype == 'D':
        IterAllButAxis[complex](_FMMdirect4_ptr, np.PyArray_Ravel(input_array, np.NPY_CORDER), np.PyArray_Ravel(output_array, np.NPY_CORDER), st, N, axis, shape, shape, &d4)
    else:
        raise NotImplementedError



#cython: boundscheck=False
#cython: wraparound=False
#cython: language_level=3

import numpy as np
cimport cython
cimport numpy as np

ctypedef fused T:
    np.float64_t
    np.complex128_t

ctypedef np.int64_t int_t

def apply_mask(u_hat, mask):
    if mask is not None:
        if u_hat.ndim == mask.ndim:
            mask = np.broadcast_to(mask, u_hat.shape).copy()
            if mask.ndim == 1:
                u_hat = apply_mask_1D(u_hat, mask)
            elif mask.ndim == 2:
                u_hat = apply_mask_2D(u_hat, mask)
            elif mask.ndim == 3:
                u_hat = apply_mask_3D(u_hat, mask)
            elif mask.ndim == 4:
                u_hat = apply_mask_4D(u_hat, mask)
            else:
                u_hat *= mask
        elif u_hat.ndim == mask.ndim + 1:
            mask = np.broadcast_to(mask, u_hat.shape[1:]).copy()
            if mask.ndim == 1:
                u_hat = apply_bmask_1D(u_hat, mask)
            elif mask.ndim == 2:
                u_hat = apply_bmask_2D(u_hat, mask)
            elif mask.ndim == 3:
                u_hat = apply_bmask_3D(u_hat, mask)
            elif mask.ndim == 4:
                u_hat = apply_bmask_4D(u_hat, mask)
            else:
                u_hat *= mask
        elif u_hat.ndim == mask.ndim + 2:
            mask = np.broadcast_to(mask, u_hat.shape[2:]).copy()
            if mask.ndim == 1:
                u_hat = apply_b2mask_1D(u_hat, mask)
            elif mask.ndim == 2:
                u_hat = apply_b2mask_2D(u_hat, mask)
            elif mask.ndim == 3:
                u_hat = apply_b2mask_3D(u_hat, mask)
            elif mask.ndim == 4:
                u_hat = apply_b2mask_4D(u_hat, mask)
            else:
                u_hat *= mask
        else:
            u_hat *= mask
    return u_hat

def apply_mask_1D(T[::1] u, int_t[::1] mask):
    cdef int i
    for i in range(mask.shape[0]):
        if mask[i] == 0:
            u[i] = 0
    return u

def apply_mask_2D(T[:, ::1] u, int_t[:, ::1] mask):
    cdef int i, j
    for i in range(mask.shape[0]):
        for j in range(mask.shape[1]):
            if mask[i, j] == 0:
                u[i, j] = 0
    return u

def apply_mask_3D(T[:, :, ::1] u, int_t[:, :, ::1] mask):
    cdef int i, j, k
    for i in range(mask.shape[0]):
        for j in range(mask.shape[1]):
            for k in range(mask.shape[2]):
                if mask[i, j, k] == 0:
                    u[i, j, k] = 0
    return u

def apply_mask_4D(T[:, :, :, ::1] u, int_t[:, :, :, ::1] mask):
    cdef int i, j, k, l
    for i in range(mask.shape[0]):
        for j in range(mask.shape[1]):
            for k in range(mask.shape[2]):
                for l in range(mask.shape[3]):
                    if mask[i, j, k, l] == 0:
                        u[i, j, k, l] = 0
    return u

def apply_bmask_1D(T[:, ::1] u, int_t[::1] mask):
    cdef int i, j
    for j in range(u.shape[1]):
        if mask[j] == 0:
            for i in range(u.shape[0]):
                u[i, j] = 0
    return u

def apply_bmask_2D(T[:, :, ::1] u, int_t[:, ::1] mask):
    cdef int i, j, k
    for j in range(mask.shape[0]):
        for k in range(mask.shape[1]):
            if mask[j, k] == 0:
                for i in range(u.shape[0]):
                    u[i, j, k] = 0
    return u

def apply_bmask_3D(T[:, :, :, ::1] u, int_t[:, :, ::1] mask):
    cdef int i, j, k, l
    for j in range(mask.shape[0]):
        for k in range(mask.shape[1]):
            for l in range(mask.shape[2]):
                if mask[j, k, l] == 0:
                    for i in range(u.shape[0]):
                        u[i, j, k, l] = 0
    return u

def apply_bmask_4D(T[:, :, :, :, ::1] u, int_t[:, :, :, ::1] mask):
    cdef int i, j, k, l, m
    for j in range(mask.shape[0]):
        for k in range(mask.shape[1]):
            for l in range(mask.shape[2]):
                for m in range(mask.shape[3]):
                    if mask[j, k, l, m] == 0:
                        for i in range(u.shape[0]):
                            u[i, j, k, l, m] = 0
    return u

def apply_b2mask_1D(T[:, :, ::1] u, int_t[::1] mask):
    cdef int i, j, k
    for k in range(u.shape[1]):
        if mask[k] == 0:
            for i in range(u.shape[0]):
                for j in range(u.shape[1]):
                    u[i, j, k] = 0
    return u

def apply_b2mask_2D(T[:, :, :, ::1] u, int_t[:, ::1] mask):
    cdef int i, j, k, l
    for k in range(mask.shape[0]):
        for l in range(mask.shape[1]):
            if mask[k, l] == 0:
                for i in range(u.shape[0]):
                    for j in range(u.shape[1]):
                        u[i, j, k, l] = 0
    return u

def apply_b2mask_3D(T[:, :, :, :, ::1] u, int_t[:, :, ::1] mask):
    cdef int i, j, k, l, m
    for k in range(mask.shape[0]):
        for l in range(mask.shape[1]):
            for m in range(mask.shape[2]):
                if mask[k, l, m] == 0:
                    for i in range(u.shape[0]):
                        for j in range(u.shape[2]):
                            u[i, j, k, l, m] = 0
    return u

def apply_b2mask_4D(T[:, :, :, :, :, ::1] u, int_t[:, :, :, ::1] mask):
    cdef int i, j, k, l, m, n
    for k in range(mask.shape[0]):
        for l in range(mask.shape[1]):
            for m in range(mask.shape[2]):
                for n in range(mask.shape[3]):
                    if mask[k, l, m, n] == 0:
                        for i in range(u.shape[0]):
                            for j in range(u.shape[1]):
                                u[i, j, k, l, m, n] = 0
    return u



import sys
import os
import numpy as np
from mpi4py import MPI
from mpi4py_fft.io import NCFile, HDF5File
from .generate_xdmf import generate_xdmf

__all__ = ['HDF5File', 'NCFile', 'ShenfunFile', 'Checkpoint', 'generate_xdmf']

comm = MPI.COMM_WORLD

def ShenfunFile(name, T, backend='hdf5', mode='r', mesh='quadrature', **kw):
    """Return a file handler

    Parameters
    ----------
    name : str
        Name of file, without ending
    T : :class:`.TensorProductSpace`
        The space used for the data stored. Can also be
        :class:`.CompositeSpace` or
        :class:`.VectorSpace`.
    backend : str, optional
        ``hdf5`` or ``netcdf4``. Default is ``hdf5``.
    mode : str, optional
        ``r`` or ``w``. Default is ``r``.
    mesh : str, optional
        - 'quadrature' - use quadrature mesh of self
        - 'uniform' - use uniform mesh for non-periodic bases

    Returns
    -------
    Class instance
        Instance of either :class:`.HDF5File` or :class:`.NCFile`
    """
    if backend.lower() == 'hdf5':
        return HDF5File(name+'.h5', domain=[np.squeeze(d) for d in T.mesh(kind=mesh)], mode=mode, **kw)
    assert kw.get('forward_output', False) is False, "NetCDF4 cannot store complex arrays, use HDF5"
    return NCFile(name+'.nc', domain=[np.squeeze(d) for d in T.mesh(kind=mesh)], mode=mode, **kw)

class Checkpoint:
    """Class for checkpointing simulations

    Checkpoint data are used to store intermediate simulation results, and can
    be used to restart a simulation at a later stage, with no loss of accuracy.

    Data is provided as dictionaries. The checkpoint dictionary is represented
    as::

        data = {
                '0': {'U': [U_hat], 'phi': [phi_hat]},
                '1': {'U': [U0_hat], 'phi': [phi_hat0]},
                 ...
                }

    which contains solutions to be stored at possibly several different timesteps.
    The current timestep is 0, previous is 1 and so on if more is needed by the
    integrator. Note that checkpoint is storing results from spectral space.

    """
    def __init__(self, filename, checkevery=10, data={}):
        self.f = None
        self.filename = filename
        self.data = data
        self.checkevery = checkevery

    def open(self, mode='r+'):
        import h5py
        self.f = h5py.File(self.filename+'.chk.h5', mode, driver="mpio", comm=comm)

    def close(self):
        if self.f:
            self.f.close()

    def update(self, t, tstep):
        if self.f is None:
            self.open(mode='w')
            self.f.attrs.create('tstep', 0)
            self.f.attrs.create('t', 0.0)
            self.close()

        kill = self.check_if_kill()
        if tstep % self.checkevery == 0 or kill:
            if comm.Get_rank() == 0: # for safety
                os.system(f'cp {self.filename}.chk.h5 {self.filename}.old.chk.h5')
            self.open()
            for key, val in self.data.items():
                self.write(int(key), val)
            self.f.attrs['tstep'] = tstep
            self.f.attrs['t'] = t
            self.close()
            if kill:
                sys.exit(1)

    def write(self, step, d):
        for name, val in d.items():
            self.f.require_group(name)
            for u in val:
                s = u.local_slice()
                self.f[name].require_dataset(str(step), shape=u.global_shape, dtype=u.dtype)
                self.f["/".join((name, str(step)))][s] = u

    def read(self, u, name, **kw):
        step = kw.get('step', 0)
        self.open()
        s = u.local_slice()
        dset = "/".join((name, str(step)))
        u[:] = self.f[dset][s]
        self.close()

    @staticmethod
    def check_if_kill():
        """Check if user has put a file named killshenfun in running folder."""
        found = 0
        if 'killshenfun' in os.listdir(os.getcwd()):
            found = 1
        collective = comm.allreduce(found)
        if collective > 0:
            if comm.Get_rank() == 0:
                os.remove('killshenfun')
                print('killshenfun Found! Stopping simulations cleanly by checkpointing...')
            return True
        else:
            return False



"""
Module for handling Fourier diagonal matrices
"""
from __future__ import division
import functools
from numbers import Number
import numpy as np
import sympy as sp
from shenfun.matrixbase import SpectralMatrix, SpectralMatDict
from shenfun import la
from . import bases

R2C = bases.R2C
C2C = bases.C2C

xp = sp.Symbol('x', real=True, positive=True)


class Acos2mat(SpectralMatrix):
    r"""Stiffness matrix for :math:`A=(a_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        a_{kj}=(\exp(i k x), \cos^2(x) \exp(i l x)'')

    where test and trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test = self.testfunction
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {0: -0.5*k**2,
             2: -0.25*k[2:]**2,
             -2: -0.25*k[:-2]**2,
             N-2: -0.25*k[-2:]**2,
             -(N-2): -0.25*k[:2]**2}
        return d


class Acosmat(SpectralMatrix):
    r"""Stiffness matrix for :math:`A=(a_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        a_{kj}=(\exp(i k x), \cos(x) \exp(i l x)'')

    where test and trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test = self.testfunction
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {1: -0.5*k[1:]**2,
             -1: -0.5*k[:-1]**2,
             N-1: -0.5*k[-1]**2}
        return d


class Csinmat(SpectralMatrix):
    r"""Derivative matrix for :math:`C=(c_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        c_{kj}=(\exp(i k x), \sin(x) \exp(i l x)')

    where test and trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test = self.testfunction
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {1: -0.5*k[1:],
             -1: 0.5*k[:-1],
             N-1: -0.5}
        return d


class Csincosmat(SpectralMatrix):
    r"""Derivative matrix for :math:`C=(c_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        c_{kj}=(\exp(i k x), \sin(2x)/2 \exp(i l x)')

    where test and trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test = self.testfunction
        k = test[0].wavenumbers(bcast=False, scaled=False, eliminate_highest_freq=False)
        N = test[0].N
        d = {2: -0.25*k[2:],
             -2: 0.25*k[:-2],
             N-2: 0.25*k[-2:],
             -(N-2): -0.25*k[:2]}
        return d

class Bcos2mat(SpectralMatrix):
    r"""Matrix for :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(\exp(i k x), \cos^2(x) \exp(i l x))

    where test and trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test = self.testfunction
        N = test[0].N
        d = {0: 0.5,
             2: 0.25,
             -2: 0.25,
             N-2: 0.25,
             -(N-2): 0.25}
        return d


class Bcosmat(SpectralMatrix):
    r"""Matrix for :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(\exp(i k x), \cos(x) \exp(i l x))

    where test and trial spaces have dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test = self.testfunction
        N = test[0].N
        d = {1: 0.5,
             -1: 0.5,
             N-1: 0.5,
             -(N-1): 0.5}
        return d

class FourierMatDict(SpectralMatDict):
    def __missing__(self, key):
        measure = 1 if len(key) == 2 else key[2]
        c = functools.partial(FourierMatrix, measure=measure)
        self[key] = c
        return c

class FourierMatrix(SpectralMatrix):
    def __init__(self, test, trial, scale=1, measure=1, assemble=None, kind=None, fixed_resolution=None):
        SpectralMatrix.__init__(self, test, trial, scale=scale, measure=measure, assemble=assemble, kind=kind, fixed_resolution=fixed_resolution)

    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        N = test[0].N
        d = {}
        if isinstance(self.measure, Number):
            k = test[0].wavenumbers(N, scaled=False)
            if isinstance(test[1], (int, np.integer)):
                k_test, k_trial = test[1], trial[1]
            elif isinstance(test[1], np.ndarray):
                assert len(test[1]) == 1
                k_test = test[1][(0,)*np.ndim(test[1])]
                k_trial = trial[1][(0,)*np.ndim(trial[1])]
            else:
                raise RuntimeError
            if abs(k_trial) + abs(k_test) > 0:
                if N % 2 == 0 and (k_trial + k_test) % 2 == 1:
                    pass
                    #k[N//2] = 0
                val = (1j*k)**(k_trial)*(-1j*k)**k_test
                if (k_trial + k_test) % 2 == 0:
                    val = val.real
                d = {0: val*float(test[0].domain_factor())}

            else:
                d = {0: float(test[0].domain_factor())}
        else:
            d = None
        return d

    def solve(self, b, u=None, axis=0, constraints=()):
        if self.measure == 1:
            N = self.shape[0]

            if u is None:
                u = b
            else:
                assert u.shape == b.shape

            with np.errstate(divide='ignore'):
                d = 1./self[0]
            if isinstance(d, np.ndarray):
                if np.isinf(d[0]):
                    d[0] = 0
                if np.isinf(d[N//2]):
                    d[N//2] = 0
                sl = [np.newaxis]*u.ndim
                sl[axis] = slice(None)
                u[:] = b*d[tuple(sl)]
            else:
                u[:] = b*d

            u /= self.scale
            return u

        return la.Solver(self)(b, u=u, axis=axis, constraints=constraints)


mat = FourierMatDict({
    ((C2C, 0), (C2C, 2), sp.cos(xp)**2): Acos2mat,
    ((C2C, 0), (C2C, 2), sp.cos(xp)): Acosmat,
    ((C2C, 0), (C2C, 1), sp.sin(xp)): Csinmat,
    ((C2C, 0), (C2C, 1), sp.sin(2*xp)/2): Csincosmat,
    ((C2C, 0), (C2C, 1), sp.sin(xp)*sp.cos(xp)): Csincosmat,
    ((C2C, 0), (C2C, 0), sp.cos(xp)**2): Bcos2mat,
    ((C2C, 0), (C2C, 0), sp.cos(xp)): Bcosmat,
})

#mat = FourierMatDict({})



r"""
Module for defining function spaces in the Legendre family.

A function is approximated in the Legendre basis as

..  math::

    u(x) = \sum_{i=0}^{N-1} \hat{u}_i L_i(x)

where :math:`L_i(x)` is the i'th Legendre polynomial of the first kind.
The Legendre polynomials are orthogonal with weight :math:`\omega=1`

.. math::

    \int_{-1}^1 L_i L_k dx = \frac{2}{2k+1} \delta_{ki}.

All other bases defined in this module are combinations of :math:`L_i`'s.
For example, a Dirichlet basis is

.. math::

    \phi_i = L_i - L_{i+2}

The basis is implemented using a stencil matrix :math:`K \in \mathbb{R}^{N-2 \times N}`,
such that

.. math::

    \boldsymbol{\phi} = K \boldsymbol{L},

where :math:`\boldsymbol{\phi}=(\phi_0, \phi_1, \ldots, \phi_{N-3})` and
:math:`\boldsymbol{L}=(L_0, L_1, \ldots, L_{N-1})`. For the Dirichlet basis
:math:`K = (\delta_{i, j} - \delta_{i+2, j})_{i,j=0}^{N-2, N}`.

The stencil matrix is used to transfer any composite basis back and forth
to the orthogonal basis.

"""

from __future__ import division
import sympy as sp
import numpy as np
from numpy.polynomial import legendre as leg
from scipy.special import eval_legendre
from mpi4py_fft import fftw
from shenfun.config import config
from shenfun.spectralbase import Transform, getCompositeBase, getBCGeneric, \
    BoundaryConditions, islicedict, slicedict
from shenfun.matrixbase import SparseMatrix
from shenfun.utilities import n
from shenfun.jacobi import JacobiBase
from .lobatto import legendre_lobatto_nodes_and_weights
from . import fastgl


bases = ['Orthogonal',
         'ShenDirichlet',
         'ShenNeumann',
         'ShenBiharmonic',
         'ShenBiPolar',
         'LowerDirichlet',
         'NeumannDirichlet',
         'DirichletNeumann',
         'UpperDirichlet',
         'UpperDirichletNeumann',
         'BeamFixedFree',
         'Generic']
bcbases = ['BCGeneric']
testbases = ['Phi1', 'Phi2', 'Phi3', 'Phi4', 'Phi6']
__all__ = bases + bcbases + testbases

#pylint: disable=method-hidden,no-else-return,not-callable,abstract-method,no-member,cyclic-import

xp = sp.Symbol('x', real=True)


class Orthogonal(JacobiBase):
    r"""Function space for a regular Legendre series

    The orthogonal basis is

    .. math::

        L_k, \quad k = 0, 1, \ldots, N-1,

    where :math:`L_k` is the :math:`k`'th Legendre polynomial.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """

    def __init__(self, N, quad="LG", domain=(-1, 1), dtype=float, padding_factor=1,
                 dealias_direct=False, coordinates=None, **kw):
        JacobiBase.__init__(self, N, quad=quad, alpha=0, beta=0, domain=domain, dtype=dtype,
                            padding_factor=padding_factor, dealias_direct=dealias_direct,
                            coordinates=coordinates)
        self.plan(int(padding_factor*N), 0, dtype, {})

    @staticmethod
    def family():
        return 'legendre'

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        if self.quad == "LG":
            points, weights = fastgl.leggauss(N)

        elif self.quad == "GL":
            points, weights = legendre_lobatto_nodes_and_weights(N)
        else:
            raise NotImplementedError

        if map_true_domain is True:
            points = self.map_true_domain(points)

        return points, weights

    def vandermonde(self, x):
        return leg.legvander(x, self.shape(False)-1)

    def get_orthogonal(self, **kwargs):
        d = dict(quad=self.quad,
                 domain=self.domain,
                 dtype=self.dtype,
                 padding_factor=self.padding_factor,
                 dealias_direct=self.dealias_direct,
                 coordinates=self.coors.coordinates)
        d.update(kwargs)
        return Orthogonal(self.N, **d)

    def orthogonal_basis_function(self, i=0, x=xp):
        return sp.legendre(i, x)

    def L2_norm_sq(self, i):
        return 2/(2*i+1)

    def l2_norm_sq(self, i=None):
        if i is None:
            f = 2/(2*np.arange(self.N)+1)
            if self.quad == 'GL':
                f[-1] = 2/(self.N-1)
            return f
        elif i == self.N-1 and self.quad == 'GL':
            return 2/(self.N-1)
        return 2/(2*i+1)

    def evaluate_basis(self, x, i=0, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape)
        output_array = eval_legendre(i, x, out=output_array)
        return output_array

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        if x is None:
            x = self.mesh(False, False)
        if output_array is None:
            output_array = np.zeros(x.shape)
        x = np.atleast_1d(x)
        basis = np.zeros(self.shape(True))
        basis[i] = 1
        basis = leg.Legendre(basis)
        if k > 0:
            basis = basis.deriv(k)
        output_array[:] = basis(x)
        return output_array

    def evaluate_basis_derivative_all(self, x=None, k=0, argument=0):
        if x is None:
            x = self.mesh(False, False)
        V = self.vandermonde(x)
        M = V.shape[-1]
        if k > 0:
            D = np.zeros((M, M))
            D[:-k] = leg.legder(np.eye(M, M), k)
            V = np.dot(V, D)
        return V

    def _evaluate_expansion_all(self, input_array, output_array, x=None, kind='fast'):

        if kind != 'fast' or self.quad != 'LG':
            JacobiBase._evaluate_expansion_all(self, input_array, output_array, x, kind=kind)
            return

        assert input_array is self.backward.tmp_array
        assert output_array is self.backward.output_array
        output_array = self.backward.xfftn()

    def _evaluate_scalar_product(self, kind='fast'):
        if kind != 'fast' or self.quad != 'LG':
            JacobiBase._evaluate_scalar_product(self, kind=kind)
            return
        out = self.scalar_product.xfftn()
        out *= 1/self.domain_factor()

    def eval(self, x, u, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.dtype)
        x = self.map_reference_domain(x)
        output_array[:] = leg.legval(x, u)
        return output_array

    @property
    def is_orthogonal(self):
        return True

    @staticmethod
    def short_name():
        return 'L'

    def stencil_matrix(self, N=None):
        N = self.N if N is None else N
        return SparseMatrix({0: 1}, (N, N))

    def sympy_stencil(self, i=sp.Symbol('i', integer=True), j=sp.Symbol('j', integer=True)):
        return sp.KroneckerDelta(i, j)

    def get_recursion_matrix(self, M, N):
        k = np.arange(max(M, N))
        return SparseMatrix({-1: (k[:min(N, M-1)]+1)/(2*k[:min(N, M-1)]+1),
                             1: (k[:min(M, N-1)]+1)/(2*k[:min(M, N-1)]+3)}, shape=(M, N))

    def get_bc_space(self):
        if self._bc_space:
            return self._bc_space
        self._bc_space = BCGeneric(self.N, bc=self.bcs, domain=self.domain)
        return self._bc_space

    def to_ortho(self, input_array, output_array=None):
        assert input_array.function_space().__class__.__name__ == 'Orthogonal'
        if output_array:
            output_array[:] = input_array
            return output_array
        return input_array

    def to_chebyshev(self, input_array, output_array=None):
        from shenfun.forms.arguments import Function, FunctionSpace
        assert input_array.function_space().__class__.__name__ == 'Orthogonal'
        C = FunctionSpace(input_array.function_space().N, 'C')
        return Function(C, buffer=self._leg2cheb(input_array, output_array))

    def plan(self, shape, axis, dtype, options):
        from .dlt import DLT, Leg2Cheb, Cheb2Leg
        from shenfun.chebyshev.bases import DCTWrap
        if shape in (0, (0,)):
            return

        if isinstance(axis, tuple):
            assert len(axis) == 1
            axis = axis[-1]

        if isinstance(self.forward, Transform):
            if self.forward.input_array.shape == shape and self.axis == axis:
                # Already planned
                return

        opts = config['fftw']['dlt']
        opts['overwrite_input'] = 'FFTW_PRESERVE_INPUT'
        opts.update(options)
        flags = (fftw.flag_dict[opts['planner_effort']],
                 fftw.flag_dict[opts['overwrite_input']])
        threads = opts['threads']
        U = fftw.aligned(shape, dtype=float)
        xfftn_fwd = DLT(U, axes=(axis,), kind='scalar product', threads=threads, flags=flags)
        V = xfftn_fwd.output_array
        xfftn_bck = DLT(V, axes=(axis,), kind='backward', threads=threads, flags=flags, output_array=U)
        V.fill(0)
        U.fill(0)
        self._leg2cheb = xfftn_fwd.leg2chebclass

        if np.dtype(dtype) is np.dtype('complex'):
            # dct only works on real data, so need to wrap it
            U = fftw.aligned(shape, dtype=complex)
            V = fftw.aligned(shape, dtype=complex)
            U.fill(0)
            V.fill(0)
            xfftn_fwd = DCTWrap(xfftn_fwd, U, V)
            xfftn_bck = DCTWrap(xfftn_bck, V, U)

        self.axis = axis
        if self.padding_factor != 1:
            trunc_array = self._get_truncarray(shape, V.dtype)
            self.scalar_product = Transform(self.scalar_product, xfftn_fwd, U, V, trunc_array)
            self.forward = Transform(self.forward, xfftn_fwd, U, V, trunc_array)
            self.backward = Transform(self.backward, xfftn_bck, trunc_array, V, U)
        else:
            self.scalar_product = Transform(self.scalar_product, xfftn_fwd, U, V, V)
            self.forward = Transform(self.forward, xfftn_fwd, U, V, V)
            self.backward = Transform(self.backward, xfftn_bck, V, V, U)

        self.si = islicedict(axis=self.axis, dimensions=U.ndim)
        self.sl = slicedict(axis=self.axis, dimensions=U.ndim)


CompositeBase = getCompositeBase(Orthogonal)
BCGeneric = getBCGeneric(CompositeBase)

class ShenDirichlet(CompositeBase):
    r"""Function space for Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_k - L_{k+2}, \, k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= \frac{1}{2}(L_0-L_1), \\
        \phi_{N-1} &= \frac{1}{2}(L_0+L_1),

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a \text{ and } u(1) = b.

    The last two basis functions are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : tuple of numbers
        Boundary conditions at edges of domain
    domain : 2-tuple of numbers, optional
        The computational domain
    scaled : bool, optional
        Whether or not to scale test functions with 1/sqrt(4k+6).
        Scaled test functions give a stiffness matrix equal to the
        identity matrix.
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 scaled=False, padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc, scaled=scaled,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 2: -1}
        if self.is_scaled():
            self._stencil = {0: 1/sp.sqrt(4*n+6), 2: -1/sp.sqrt(4*n+6)}

    @staticmethod
    def boundary_condition():
        return 'Dirichlet'

    @staticmethod
    def short_name():
        return 'SD'


class Phi1(CompositeBase):
    r"""Function space for Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= \frac{1}{2}(L_k - L_{k+2}) = \frac{(2k+3)(1-x^2)}{2(k+1)(k+2)} L'_{k+1}, \, k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= \frac{1}{2}(L_0-L_1), \\
        \phi_{N-1} &= \frac{1}{2}(L_0+L_1),

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a \text{ and } u(1) = b.

    The last two basis functions are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : tuple of numbers
        Boundary conditions at edges of domain
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: sp.S.Half, 2: -sp.S.Half}

    @staticmethod
    def boundary_condition():
        return 'Dirichlet'

    @staticmethod
    def short_name():
        return 'P1'


class ShenNeumann(CompositeBase):
    r"""Function space for Neumann boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_{k} -  \frac{k(k+1)}{(k+2)(k+3)}L_{k+2}, \, k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= \frac{1}{6}(3L_1-L_2), \\
        \phi_{N-1} &= \frac{1}{6}(3L_1+L_2),

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u'(-1) &= a \text{ and } u'(1) = b.

    The last two bases are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : 2-tuple of numbers
        Boundary conditions at edges of domain
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0), domain=(-1, 1), padding_factor=1,
                 dealias_direct=False, dtype=float, coordinates=None, **kw):
        if isinstance(bc, (tuple, list)):
            bc = BoundaryConditions({'left': {'N': bc[0]}, 'right': {'N': bc[1]}}, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 2: -n*(n + 1)/(n**2 + 5*n + 6)}

    @staticmethod
    def boundary_condition():
        return 'Neumann'

    @staticmethod
    def short_name():
        return 'SN'


class ShenBiharmonic(CompositeBase):
    r"""Function space for biharmonic equation

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_n - \frac{4n+10}{2n+7}L_{n+2}+\frac{2 n + 3}{2 n + 7}L_{n+4}, \, k=0, 1, \ldots, N-5, \\
        \phi_{N-4} &= \tfrac{1}{2}L_0-\tfrac{3}{5}L_1+\tfrac{1}{10}L_3, \\
        \phi_{N-3} &= \tfrac{1}{6}L_0-\tfrac{1}{10}L_1-\tfrac{1}{6}L_2+\tfrac{1}{10}L_3, \\
        \phi_{N-2} &= \tfrac{1}{2}L_0+\tfrac{3}{5}L_1-\tfrac{1}{10}L_3), \\
        \phi_{N-1} &= -\tfrac{1}{6}L_0-\tfrac{1}{10}L_1+\tfrac{1}{6}L_2+\tfrac{1}{10}L_3,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1)&=a, u'(-1) = b, u(1)=c, u'(1) = d.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 4-tuple of numbers, optional
        The values of the 4 boundary conditions at x=(-1, 1).
        The two conditions on x=-1 first, and then x=1.
        With (a, b, c, d) corresponding to
        bc = {'left': [('D', a), ('N', b)], 'right': [('D', c), ('N', d)]}
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0, 0, 0), domain=(-1, 1), padding_factor=1,
                 dealias_direct=False, dtype=float, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 2: -(4*n + 10)/(2*n + 7), 4: (2*n + 3)/(2*n + 7)}

    @staticmethod
    def boundary_condition():
        return 'Biharmonic'

    @staticmethod
    def short_name():
        return 'SB'


class Phi2(CompositeBase):
    r"""Function space for biharmonic equation

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-5` are

    .. math::

        \phi_k &= \frac{(1-x^2)^2 L''_{k+2}}{h^{(2)}_{k+2}}, \\
        h^{(2)}_{k+2} &= \int_{-1}^1 L''_{k+2} L''_{k+2} (1-x^2)^2 dx, \\
               &= \frac{2 (k+1)(k+2)(k+3)(k+4)}{2k+5},

    which (along with boundary functions) becomes the basis

    .. math::

        \phi_k &= \frac{1}{2(2k+3)}\left(L_k - \frac{2(2k+5)}{2k+7}L_{k+2} + \frac{2k+3}{2k+7}L_{k+4}\right), \, k=0, 1, \ldots, N-5, \\
        \phi_{N-4} &= \tfrac{1}{2}L_0-\tfrac{3}{5}L_1+\tfrac{1}{10}L_3, \\
        \phi_{N-3} &= \tfrac{1}{6}L_0-\tfrac{1}{10}L_1-\tfrac{1}{6}L_2+\tfrac{1}{10}L_3, \\
        \phi_{N-2} &= \tfrac{1}{2}L_0+\tfrac{3}{5}L_1-\tfrac{1}{10}L_3, \\
        \phi_{N-1} &= -\tfrac{1}{6}L_0-\tfrac{1}{10}L_1+\tfrac{1}{6}L_2+\tfrac{1}{10}L_3,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1)&=a, u'(-1) = b, u(1)=c, u'(1) = d.

    The last four basis functions are for boundary conditions and only used if a, b, c or d are
    different from 0. In one dimension :math:`\hat{u}_{N-4}=a`, :math:`\hat{u}_{N-3}=b`,
    :math:`\hat{u}_{N-2}=c` and :math:`\hat{u}_{N-1}=d`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 4-tuple of numbers, optional
        The values of the 4 boundary conditions at x=(-1, 1).
        The two on x=-1 first and then x=1. (a, b, c, d)
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0, 0, 0), domain=(-1, 1), padding_factor=1,
                 dealias_direct=False, dtype=float, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1/(2*(2*n+3)), 2: -(2*n+5)/(2*n+7)/(2*n+3), 4: 1/(2*(2*n+7))}

    @staticmethod
    def boundary_condition():
        return 'Biharmonic'

    @staticmethod
    def short_name():
        return 'P2'


class Phi3(CompositeBase):
    r"""Function space for 6th order equations

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-7` are

    .. math::
        \phi_k &= \frac{(1-x^2)^3}{h^{(3)}_{k+3}} L^{(3)}_{k+3}}, \, k=0, 1, \ldots, N-7, \\
        h^{(3)}_{k+3} &= \frac{2\Gamma(k+7)}{\Gamma(k+1)(2k+7)} = \int_{-1}^1 L^{(3)}_{k+3} L^{(3)}_{k+3}(1-x^2)^3 dx,

    where :math:`L^{(3)}_k` is the 3'rd derivative of :math:`L_k`.
    The 6 boundary basis functions are computed using :func:`.jacobi.findbasis.get_bc_basis`,
    but they are too messy to print here. We have

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a, u'(-1)=b, u''(-1)=c, u(1)=d u'(1)=e, u''(1)=f.

    The last 6 basis functions are for boundary conditions and only used if there
    are nonzero boundary conditions.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature
        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 6-tuple of numbers, optional
        Boundary conditions.
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="LG", bc=(0,)*6, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        #self._stencil = {
        #    0: sp.simplify(matpow(b, 3, 0, 0, n+3, n) / h(0, 0, n, 0)),
        #    2: sp.simplify(matpow(b, 3, 0, 0, n+3, n+2) / h(0, 0, n+2, 0)),
        #    4: sp.simplify(matpow(b, 3, 0, 0, n+3, n+4) / h(0, 0, n+4, 0)),
        #    6: sp.simplify(matpow(b, 3, 0, 0, n+3, n+6) / h(0, 0, n+6, 0))}
        self._stencil = {
            0: 1/(2*(4*n**2 + 16*n + 15)),
            2: -3/(8*n**2 + 48*n + 54),
            4: 3/(2*(4*n**2 + 32*n + 55)),
            6: -1/(8*n**2 + 80*n + 198)
        }

    @staticmethod
    def boundary_condition():
        return '6th order'

    @staticmethod
    def short_name():
        return 'P3'


class Phi4(CompositeBase):
    r"""Function space with 2 Dirichlet and 6 Neumann boundary conditions

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-9` are

    .. math::

        \phi_k &= \frac{(1-x^2)^4}{h^{(4)}_{k+4}} L^{(4)}_{k+4}, \\
        h^{(4)}_{k+4} &= \frac{2\Gamma(k+9)}{\Gamma(k+1)(2k+9)} = \int_{-1}^1 L^{(4)}_{k+4} L^{(4)}_{k+4} (1-x^2)^4 dx,

    where :math:`L^{(4)}_k` is the 4'th derivative of :math:`L_k`.
    The boundary basis for inhomogeneous boundary conditions is too
    messy to print, but can be obtained using :func:`~shenfun.utilities.findbasis.get_bc_basis`.
    We have

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a, u'(-1)=b, u''(-1)=c, u'''(-1)=d, u(1)=e u'(1)=f, u''(1)=g, u'''(1)=h.

    The last 8 basis functions are for boundary conditions and only used if there
    are nonzero boundary conditions.

    Parameters
    ----------
    N : int, optional
        Number of quadrature points
    quad : str, optional
        Type of quadrature
        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 8-tuple of numbers
    domain : 2-tuple of numbers, optional
        The computational domain
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0,)*8, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        #self._stencil = {
        #   0: sp.simplify(matpow(b, 4, 0, 0, n+4, n) / h(0, 0, n, 0)),
        #   2: sp.simplify(matpow(b, 4, 0, 0, n+4, n+2) / h(0, 0, n+2, 0)),
        #   4: sp.simplify(matpow(b, 4, 0, 0, n+4, n+4) / h(0, 0, n+4, 0)),
        #   6: sp.simplify(matpow(b, 4, 0, 0, n+4, n+6) / h(0, 0, n+6, 0)),
        #   8: sp.simplify(matpow(b, 4, 0, 0, n+4, n+8) / h(0, 0, n+8, 0))}
        # Below are the same but faster since already simplified
        self._stencil = {
            0: 1/(2*(8*n**3 + 60*n**2 + 142*n + 105)),
            2: -2/(8*n**3 + 84*n**2 + 262*n + 231),
            4: 3*(2*n + 9)/((2*n + 5)*(2*n + 7)*(2*n + 11)*(2*n + 13)),
            6: -2/(8*n**3 + 132*n**2 + 694*n + 1155),
            8: 1/(2*(8*n**3 + 156*n**2 + 1006*n + 2145))
        }

    @staticmethod
    def boundary_condition():
        return 'Biharmonic*2'

    @staticmethod
    def short_name():
        return 'P4'

class Phi6(CompositeBase):
    r"""Function space for 12th order equation

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-9` are

    .. math::

        \phi_k &= \frac{(1-x^2)^6}{h^{(6)}_{k+6}} L^{(6)}_{k+6}, \\
        h^{(6)}_{k+6} &= \int_{-1}^1 L^{(6)}_{k+6} L^{(6)}_{k+6} (1-x^2)^6 dx,

    where :math:`L^{(6)}_k` is the 6'th derivative of :math:`L_k`.
    The boundary basis for inhomogeneous boundary conditions is too
    messy to print, but can be obtained using :func:`~shenfun.utilities.findbasis.get_bc_basis`.

    Parameters
    ----------
    N : int, optional
        Number of quadrature points
    quad : str, optional
        Type of quadrature
        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 12-tuple of numbers
    domain : 2-tuple of numbers, optional
        The computational domain
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0,)*12, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        #self._stencil = {
        #   0: sp.simplify(matpow(b, 6, 0, 0, n+6, n) / h(0, 0, n, 0)),
        #   2: sp.simplify(matpow(b, 6, 0, 0, n+6, n+2) / h(0, 0, n+2, 0)),
        #   4: sp.simplify(matpow(b, 6, 0, 0, n+6, n+4) / h(0, 0, n+4, 0)),
        #   6: sp.simplify(matpow(b, 6, 0, 0, n+6, n+6) / h(0, 0, n+6, 0)),
        #   8: sp.simplify(matpow(b, 6, 0, 0, n+6, n+8) / h(0, 0, n+8, 0)),
        #  10: sp.simplify(matpow(b, 6, 0, 0, n+6, n+10) / h(0, 0, n+10, 0)),
        #  12: sp.simplify(matpow(b, 6, 0, 0, n+6, n+12) / h(0, 0, n+12, 0))}
        # Below are the same but faster since already simplified
        self._stencil = {
            0: 1/(2*(2*n + 3)*(2*n + 5)*(2*n + 7)*(2*n + 9)*(2*n + 11)),
            2: -3/((2*n + 3)*(2*n + 7)*(2*n + 9)*(2*n + 11)*(2*n + 15)),
            4: 15/(2*(2*n + 5)*(2*n + 7)*(2*n + 11)*(2*n + 15)*(2*n + 17)),
            6: -10*(2*n + 13)/((2*n + 7)*(2*n + 9)*(2*n + 11)*(2*n + 15)*(2*n + 17)*(2*n + 19)),
            8: 15/(2*(2*n + 9)*(2*n + 11)*(2*n + 15)*(2*n + 19)*(2*n + 21)),
            10: -3/((2*n + 11)*(2*n + 15)*(2*n + 17)*(2*n + 19)*(2*n + 23)),
            12: 1/(2*(2*n + 15)*(2*n + 17)*(2*n + 19)*(2*n + 21)*(2*n + 23))
        }

    @staticmethod
    def boundary_condition():
        return '12th order'

    @staticmethod
    def short_name():
        return 'P6'


class BeamFixedFree(CompositeBase):
    r"""Function space for fixed free beams

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_k + a_kL_{k+1} + b_k L_{k+2} + c_k L_{k+3} + d_k L_{k+4} , \, k=0, 1, \ldots, N-5, \\
        \phi_{N-4} &= \tfrac{1}{2}L_0-\tfrac{3}{5}L_1+\tfrac{1}{10}L_3, \\
        \phi_{N-3} &= \tfrac{1}{6}L_0-\tfrac{1}{10}L_1-\tfrac{1}{6}L_2+\tfrac{1}{10}L_3, \\
        \phi_{N-2} &= \tfrac{1}{2}L_0+\tfrac{3}{5}L_1-\tfrac{1}{10}L_3), \\
        \phi_{N-1} &= -\tfrac{1}{6}L_0-\tfrac{1}{10}L_1+\tfrac{1}{6}L_2+\tfrac{1}{10}L_3,

    where

    .. math::

        a_k &= \frac{4 \left(2 n + 3\right)}{\left(n + 3\right)^{2}}, \\
        b_k &= -\frac{2 \left(n - 1\right) \left(n + 1\right) \left(n + 6\right) \left(2 n + 5\right)}{\left(n + 3\right)^{2} \left(n + 4\right) \left(2 n + 7\right)}, \\
        c_k &= -\frac{4 \left(n + 1\right)^{2} \left(2 n + 3\right)}{\left(n + 3\right)^{2} \left(n + 4\right)^{2}}, \\
        d_k &= \frac{\left(n + 1\right)^{2} \left(n + 2\right)^{2} \left(2 n + 3\right)}{\left(n + 3\right)^{2} \left(n + 4\right)^{2} \left(2 n + 7\right)}.

    We have

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1)&=a, u'(-1) = b, u''(1)=c, u'''(1) = d.

    The last four basis functions are for boundary conditions and only used if a, b, c or d are
    different from 0. In one dimension :math:`\hat{u}_{N-4}=a`, :math:`\hat{u}_{N-3}=b`,
    :math:`\hat{u}_{N-2}=c` and :math:`\hat{u}_{N-1}=d`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    4-tuple of numbers, optional
        The values of the 4 boundary conditions
        u(-1) = a, u'(-1) = b, u''(1) = c, u'''(1) = d

    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0, 0, 0), domain=(-1, 1), padding_factor=1,
                 dealias_direct=False, dtype=float, coordinates=None, **kw):
        if isinstance(bc, (tuple, list)):
            bc = BoundaryConditions({'left': {'D': bc[0], 'N': bc[1]}, 'right': {'N2': bc[2], 'N3': bc[3]}}, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {
            0: 1,
            1: 4*(2*n + 3)/(n**2 + 6*n + 9),
            2: 2*(-2*n**4 - 17*n**3 - 28*n**2 + 17*n + 30)/(2*n**4 + 27*n**3 + 136*n**2 + 303*n + 252),
            3: -(8*n**3 + 28*n**2 + 32*n + 12)/(n**4 + 14*n**3 + 73*n**2 + 168*n + 144),
            4: (2*n**5 + 15*n**4 + 44*n**3 + 63*n**2 + 44*n + 12)/(2*n**5 + 35*n**4 + 244*n**3 + 847*n**2 + 1464*n + 1008)
        }

    @staticmethod
    def boundary_condition():
        return 'BeamFixedFree'

    @staticmethod
    def short_name():
        return 'BF'


class UpperDirichlet(CompositeBase):
    r"""Function space with single Dirichlet on upper edge

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_{k} - L_{k+1}, \, k=0, 1, \ldots, N-2, \\
        \phi_{N-1} &= L_0,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(1) &= a.

    The last basis function is for boundary condition and only used if a is
    different from 0. In one dimension :math:`\hat{u}_{N-1}=a`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 2-tuple of (None, number), optional
        The number is the boundary condition value
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(None, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: -1}

    @staticmethod
    def boundary_condition():
        return 'UpperDirichlet'

    @staticmethod
    def short_name():
        return 'UD'


class ShenBiPolar(CompositeBase):
    r"""Function space for the Biharmonic equation

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= (1-x^2)^2 L'_{k+1}, \quad k=0,1, \ldots, N-5, \\
        \phi_{N-4} &= \tfrac{1}{2}L_0-\tfrac{3}{5}L_1+\tfrac{1}{10}L_3, \\
        \phi_{N-3} &= \tfrac{1}{6}L_0-\tfrac{1}{10}L_1-\tfrac{1}{6}L_2+\tfrac{1}{10}L_3, \\
        \phi_{N-2} &= \tfrac{1}{2}L_0+\tfrac{3}{5}L_1-\tfrac{1}{10}L_3), \\
        \phi_{N-1} &= -\tfrac{1}{6}L_0-\tfrac{1}{10}L_1+\tfrac{1}{6}L_2+\tfrac{1}{10}L_3,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1)&=a, u'(-1) = b, u(1)=c, u'(1) = d.

    The last four bases are for boundary conditions and only used if a, b, c or d are
    different from 0. In one dimension :math:`\hat{u}_{N-4}=a`, :math:`\hat{u}_{N-3}=b`,
    :math:`\hat{u}_{N-2}=c` and :math:`\hat{u}_{N-1}=d`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 4-tuple of numbers, optional
        The values of the 4 boundary conditions at x=(-1, 1).
        The two on x=-1 first and then x=1. (a, b, c, d)
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", domain=(-1, 1), bc=(0, 0, 0, 0), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)

    @staticmethod
    def boundary_condition():
        return 'Biharmonic'

    @staticmethod
    def short_name():
        return 'SP'

    def sympy_stencil(self, i=sp.Symbol('i', integer=True), j=sp.Symbol('j', integer=True)):
        return RuntimeError, "Not possible for current basis"

    def stencil_matrix(self, N=None):
        N = self.N if N is None else N
        k = np.arange(N)
        ak = k*(k-1)/(2*k-1)/(2*k+1)
        bk = -2*(k**2+k-1)/(2*k-1)/(2*k+3)
        ck = (k+1)*(k+2)/(2*k+1)/(2*k+3)
        d = np.zeros(N)
        d[:-4] = (k[:-4]+1)*(k[:-4]+2)/(2*k[:-4]+3)*(ak[2:-2]-bk[:-4])
        d[0] = 8/15
        d[1] = 24/35
        dm2 = np.zeros(N-2)
        dm2[:-4] = -(k[2:-4]+1)*(k[2:-4]+2)/(2*k[2:-4]+3)*ak[2:-4]
        dp2 = np.zeros(N-2)
        dp2[:-2] = (k[:-4]+1)*(k[:-4]+2)/(2*k[:-4]+3)*(bk[2:-2]-ck[:-4])
        dp2[0] = -16/21
        dp2[1] = -16/15
        dp4 = np.zeros(N-4)
        dp4[:] = (k[:-4]+1)*(k[:-4]+2)/(2*k[:-4]+3)*ck[2:-2]
        dp4[0] = 8/35
        dp4[1] = 8/21
        return SparseMatrix({-2: dm2, 0: d, 2: dp2, 4: dp4}, (N, N))

class DirichletNeumann(CompositeBase):
    r"""Function space for mixed Dirichlet/Neumann boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_{k} + \frac{2n+3}{\left(n+2\right)^{2}}L_{k+1} - \frac{\left(n+1\right)^{2}}{\left(n+2\right)^{2}} L_{k+2}, \, k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= L_0, \\
        \phi_{N-1} &= L_0+L_1,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a, u'(1)=b.

    The last two bases are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : tuple of numbers
        Boundary conditions at edges of domain. Dirichlet first.
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        if isinstance(bc, (tuple, list)):
            bc = BoundaryConditions({'left': {'D': bc[0]}, 'right': {'N': bc[1]}}, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: (2*n + 3)/(n**2 + 4*n + 4), 2: -(n**2 + 2*n + 1)/(n**2 + 4*n + 4)}

    @staticmethod
    def boundary_condition():
        return 'DirichletNeumann'

    @staticmethod
    def short_name():
        return 'DN'


class LowerDirichlet(CompositeBase):
    r"""Function space with single Dirichlet boundary condition

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_{k} + L_{k+1}, \, k=0, 1, \ldots, N-2, \\
        \phi_{N-1} &= L_0,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a.

    The last basis function is for boundary condition and only used if a is
    different from 0. In one dimension :math:`\hat{u}_{N-1}=a`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : tuple of (number, None)
        Boundary conditions at edges of domain.
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, None), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: 1}

    @staticmethod
    def boundary_condition():
        return 'LowerDirichlet'

    @staticmethod
    def short_name():
        return 'LD'


class NeumannDirichlet(CompositeBase):
    r"""Function space for mixed Neumann/Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_{k} - \frac{2n+3}{\left(n+2\right)^{2}}L_{k+1} - \frac{\left(n+1\right)^{2}}{\left(n+2\right)^{2}}L_{k+2}, \, k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= -L_0+L_1, \\
        \phi_{N-1} &= L_0,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u'(-1) &= a, u(1)=b.

    The last two bases are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : tuple of numbers
        Boundary conditions at edges of domain. Neumann first.
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        if isinstance(bc, (tuple, list)):
            bc = BoundaryConditions({'left': {'N': bc[0]}, 'right': {'D': bc[1]}}, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: -(2*n + 3)/(n**2 + 4*n + 4), 2: -(n**2 + 2*n + 1)/(n**2 + 4*n + 4)}

    @staticmethod
    def boundary_condition():
        return 'NeumannDirichlet'

    @staticmethod
    def short_name():
        return 'ND'


class UpperDirichletNeumann(CompositeBase):
    r"""Function space for both Dirichlet and Neumann boundary conditions
    on the right hand side.

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= L_{k} - \frac{2k+3}{k+2}L_{k+1} + \frac{k+1}{k+2}L_{k+2}, \, k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= L_0, \\
        \phi_{N-1} &= -L_0+L_1,

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(1) &= a, u'(1)=b.

    The last two bases are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : tuple of numbers
        Boundary conditions at edges of domain, Dirichlet first.
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    Note
    ----
    This basis is not recommended as it leads to a poorly conditioned
    stiffness matrix.
    """
    def __init__(self, N, quad="LG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        if isinstance(bc, (tuple, list)):
            bc = BoundaryConditions({'right': {'D': bc[0], 'N': bc[1]}}, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: -(2*n + 3)/(n + 2), 2: (n + 1)/(n + 2)}

    @staticmethod
    def boundary_condition():
        return 'UpperDirichletNeumann'

    @staticmethod
    def short_name():
        return 'UDN'

class Compact3(CompositeBase):
    r"""Function space for 6th order equations

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-7` are

    .. math::
        \phi_k &= \frac{h_k}{b^{(3)}_{k+3,k}}\frac{(1-x^2)^3}{h^{(3)}_{k+3}} L^{(3)}_{k+3}}, \, k=0, 1, \ldots, N-7, \\
        h^{(3)}_{k+3} &= \frac{2\Gamma(k+7)}{\Gamma(k+1)(2k+7)} = \int_{-1}^1 L^{(3)}_{k+3} L^{(3)}_{k+3}(1-x^2)^3 dx,

    where :math:`L^{(3)}_k` is the 3'rd derivative of :math:`L_k`.
    The 6 boundary basis functions are computed using :func:`.jacobi.findbasis.get_bc_basis`,
    but they are too messy to print here. We have

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a, u'(-1)=b, u''(-1)=c, u(1)=d u'(1)=e, u''(1)=f.

    The last 6 basis functions are for boundary conditions and only used if there
    are nonzero boundary conditions.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature
        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto
    bc : 6-tuple of numbers, optional
        Boundary conditions.
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="LG", bc=(0,)*6, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)
        #self._stencil = {
        #    0: 1,
        #    2: sp.simplify(matpow(b, 3, 0, 0, n+3, n+2) / matpow(b, 3, 0, 0, n+3, n) * h(0, 0, n, 0) / h(0, 0, n+2, 0)),
        #    4: sp.simplify(matpow(b, 3, 0, 0, n+3, n+4) / matpow(b, 3, 0, 0, n+3, n) * h(0, 0, n, 0) / h(0, 0, n+4, 0)),
        #    6: sp.simplify(matpow(b, 3, 0, 0, n+3, n+6) / matpow(b, 3, 0, 0, n+3, n) * h(0, 0, n, 0) / h(0, 0, n+6, 0))}
        self._stencil = {
            0: 1,
            2: -(6*n + 15)/(2*n + 9),
            4: 3*(2*n + 3)/(2*n + 11),
            6: -(2*n + 3)*(2*n + 5)/((2*n + 9)*(2*n + 11))
        }

    @staticmethod
    def boundary_condition():
        return '6th order'

    @staticmethod
    def short_name():
        return 'C3'

class Generic(CompositeBase):
    r"""Function space for space with any boundary conditions

    Any combination of Dirichlet and Neumann is possible.

    Parameters
    ----------
    N : int, optional
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Legendre-Gauss
        - GL - Legendre-Gauss-Lobatto

    bc : dict, optional
        The dictionary must have keys 'left' and 'right', to describe boundary
        conditions on the left and right boundaries. Specify Dirichlet on both
        ends with

            {'left': {'D': a}, 'right': {'D': b}}

        for some values `a` and `b`, that will be neglected in the current
        function. Specify mixed Neumann and Dirichlet as

            {'left': {'N': a}, 'right': {'N': b}}

        For both conditions on the right do

            {'right': {'N': a, 'D': b}}

        Any combination should be possible, and it should also be possible to
        use second derivatives `N2`. See :class:`~shenfun.spectralbase.BoundaryConditions`.
    domain : 2-tuple of numbers, optional
        The computational domain
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    Note
    ----
    A test function is always using homogeneous boundary conditions.

    """
    def __init__(self, N, quad="LG", bc={}, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        if not isinstance(bc, BoundaryConditions):
            bc = BoundaryConditions(bc, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)

    @staticmethod
    def boundary_condition():
        return 'Generic'

    @staticmethod
    def short_name():
        return 'GL'



import numpy as np
from shenfun.matrixbase import SpectralMatrix, SpectralMatDict
from shenfun.optimization.cython import Matvec
from shenfun.la import TDMA
from . import bases

H = bases.Orthogonal


class BHHmat(SpectralMatrix):
    r"""Mass matrix :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(H_j, H_k)_w,

    :math:`H_k \in` :class:`.hermite.bases.Orthogonal` and test and trial spaces have
    dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], H)
        assert isinstance(trial[0], H)
        return {0: 1}

    def solve(self, b, u=None, axis=0, constraints=()):
        if u is not None:
            u[:] = b
            u /= (self.scale*self[0])
            return u

        else:
            b /= (self.scale*self[0])
            return b

    def matvec(self, v, c, format=None, axis=0):
        M = self.shape[1]
        ss = [slice(None)]*len(v.shape)
        ss[self.axis] = slice(0, M)
        c[tuple(ss)] = v
        self.scale_array(c, self.scale*self[0])
        return c


class AHHmat(SpectralMatrix):
    r"""Stiffness matrix :math:`A=(a_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        a_{kj}=(H'_j, H'_k)_w,

    :math:`H_k \in` :class:`.hermite.bases.Orthogonal` and test and trial spaces have
    dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], H)
        assert isinstance(trial[0], H)
        N = test[0].N
        k = np.arange(N, dtype=float)
        d = {0: k+0.5,
             2: -np.sqrt((k[:-2]+1)*(k[:-2]+2))/2}
        d[0][-1] = (N-1)/2.
        d[-2] = d[2].copy()
        return d

    def get_solver(self):
        return TDMA

    def matvec(self, v, c, format='cython', axis=0):
        N, M = self.shape
        c.fill(0)
        if format == 'cython' and v.ndim == 3:
            ld = self[-2]*np.ones(M-2)
            Matvec.Tridiagonal_matvec3D_ptr(v, c, ld, self[0], ld, axis)
            self.scale_array(c, self.scale)
        elif format == 'cython' and v.ndim == 2:
            ld = self[-2]*np.ones(M-2)
            Matvec.Tridiagonal_matvec2D_ptr(v, c, ld, self[0], ld, axis)
            self.scale_array(c, self.scale)
        elif format == 'cython' and v.ndim == 1:
            ld = self[-2]*np.ones(M-2)
            Matvec.Tridiagonal_matvec(v, c, ld, self[0], ld)
            self.scale_array(c, self.scale)
        elif format == 'self':
            if axis > 0:
                v = np.moveaxis(v, axis, 0)
                c = np.moveaxis(c, axis, 0)
            s = (slice(None),)+(np.newaxis,)*(v.ndim-1) # broadcasting
            c[:(N-2)] = self[2][s]*v[2:N]
            c[:N] += self[0][s]*v[:N]
            c[2:N] += self[-2][s]*v[:(N-2)]
            if axis > 0:
                v = np.moveaxis(v, 0, axis)
                c = np.moveaxis(c, 0, axis)
            self.scale_array(c, self.scale)

        else:
            format = None if format in self._matvec_methods else format
            c = super(AHHmat, self).matvec(v, c, format=format, axis=axis)

        return c


mat = SpectralMatDict({
    ((H, 0), (H, 0)): BHHmat,
    ((H, 1), (H, 1)): AHHmat
    })
